/*

jeskola ninja delay

TODO: 
	- use maxdelay attribute
	- free buffers
	- should we allocate/deallocate when adding/removing tracks?
	- nicer descriptions

*/

#include "ninja.h"
#include <lunar/fx.hpp>
#include <lunar/dsp.h>
#include "dsplib.h"

double const SilentEnough = log(1.0 / 32768);

#define MAX_TAPS		8

#define UNIT_TICK		0
#define UNIT_MS			1
#define UNIT_SAMPLE		2
#define UNIT_256		3

#define FILTER_LP		0
#define FILTER_HP		1
#define FILTER_BP		2
#define FILTER_BR		3

#define MAX_BUFFER_LENGTH 256

class CTrack
{
public:
	CBWState Filter;
	float *Buffer;
	int Length;
	int Pos;
	float Feedback;
	float WetOut;
	int Unit;

	int FilterType;
	int FilterCutoff;
	int FilterBandwidth;
};
 


class mi : public lunar::fx<mi>
{
public:
	mi() { }

	void CreateTrack(CTrack& t) {
		t.Buffer = 0;
		t.Unit = UNIT_TICK;
		t.Length = transport->samples_per_tick * 3;
		t.Pos = 0;
		t.Feedback = 0.3f;
		t.WetOut = 0;
		t.FilterType = 1;
		t.FilterCutoff = 1000;
		t.FilterBandwidth = 1000;
		InitTrack(t);
		
		DSP_BW_Reset(t.Filter);
		SetFilter(t);
	}
	
	void init() {
		DSP_Init(transport->samples_per_second);
		DryThru = true;
	
		int attribute_max_delay = 1000;
		MaxDelay = (int)(transport->samples_per_second * ((float)attribute_max_delay / 1000.0));

		for (int c = 0; c < MAX_TAPS; c++)
		{
			CreateTrack(TracksL[c]);
			CreateTrack(TracksR[c]);
		}
	
		TracksL[0].WetOut = 0.3f;	// enable first track
		TracksR[0].WetOut = 0.3f;	// enable first track
	
		IdleMode = true;
		IdleCount = 0;
	}

	void process_events() {
		for (int c = 0; c < track_count; c++) {
			TickTrack(&TracksL[c], tracks[c]);
			TickTrack(&TracksR[c], tracks[c]);
		}
	
		// find max delay time slow we know when to stop wasting CPU time
		int maxdt = 0;
		for (int c = 0; c < track_count; c++)
		{
			int dt = TracksL[c].Length + (int)(SilentEnough / log(abs(TracksL[c].Feedback)) * TracksL[c].Length); 
			if (dt > maxdt)
				maxdt = dt;

			dt = TracksR[c].Length + (int)(SilentEnough / log(abs(TracksR[c].Feedback)) * TracksR[c].Length); 
			if (dt > maxdt)
				maxdt = dt;
		}
		
		DelayTime = maxdt;
		if (globals->drythru)
			DryThru = globals->drythru != 0;
	}
	
	void process_stereo(float *inL, float *inR, float *outL, float *outR, int sampleFrames) {
		Work(inL, sampleFrames, WM_READWRITE, TracksL);
		memcpy(outL, inL, sampleFrames * sizeof(float));

		Work(inR, sampleFrames, WM_READWRITE, TracksR);
		memcpy(outR, inR, sampleFrames * sizeof(float));
	}
	
	#pragma optimize ("a", on) 
	
	static void DoWork(float *pin, float *pout, float *pbuf, int c, double const wetout, double const feedback, CBWState &s)
	{
		do
		{
	
			double delay = *pbuf;
			
			double const o = s.a[0] * delay + s.a[1] * s.i[0] + s.a[2] * s.i[1] - s.a[3] * s.o[0] - s.a[4] * s.o[1];
			s.o[1] = s.o[0];
			s.o[0] = (float)o;
			s.i[1] = s.i[0];
			s.i[0] = (float)delay;
	
			*pbuf++ = (float)(feedback * o + *pin++);
			*pout++ += (float)(o * wetout);
	
		} while(--c);
	}
	
	static void DoWorkNoFB(float *pin, float *pout, float *pbuf, int c, double const wetout, CBWState &s)
	{
		do
		{
	
			double delay = *pbuf;
	
			double const o = s.a[0] * delay + s.a[1] * s.i[0] + s.a[2] * s.i[1] - s.a[3] * s.o[0] - s.a[4] * s.o[1];
			s.o[1] = s.o[0];
			s.o[0] = (float)o;
			s.i[1] = s.i[0];
			s.i[0] = (float)delay;
			
			*pbuf++ = (float)(*pin++);
			*pout++ += (float)(o * wetout);
	
		} while(--c);
	}
	
	static void DoWorkNoInput(float *pout, float *pbuf, int c, double const wetout, double const feedback, CBWState &s)
	{
		do
		{
			double delay = *pbuf;
	
			double const o = s.a[0] * delay + s.a[1] * s.i[0] + s.a[2] * s.i[1] - s.a[3] * s.o[0] - s.a[4] * s.o[1];
			s.o[1] = s.o[0];
			s.o[0] = (float)o;
			s.i[1] = s.i[0];
			s.i[0] = (float)delay;
			
			*pbuf++ = (float)(feedback * o);
			*pout++ += (float)(o * wetout);
	
		} while(--c);
	}
	
	
	
	static void DoWorkNoInputNoFB(float *pout, float *pbuf, int c, double const wetout, CBWState &s)
	{
	
		do
		{
			double delay = *pbuf;
	
			double const o = s.a[0] * delay + s.a[1] * s.i[0] + s.a[2] * s.i[1] - s.a[3] * s.o[0] - s.a[4] * s.o[1];
			s.o[1] = s.o[0];
			s.o[0] = (float)o;
			s.i[1] = s.i[0];
			s.i[0] = (float)delay;
			
			*pbuf++ = 0;
			*pout++ += (float)(o * wetout);
		} while(--c);
	}
	
	static void DoWorkNoInputNoOutput(float *pbuf, int c, double const feedback, CBWState &s)
	{
		do
		{
			double delay = *pbuf;
	
			double const o = s.a[0] * delay + s.a[1] * s.i[0] + s.a[2] * s.i[1] - s.a[3] * s.o[0] - s.a[4] * s.o[1];
			s.o[1] = s.o[0];
			s.o[0] = (float)o;
			s.i[1] = s.i[0];
			s.i[0] = (float)delay;
	
			*pbuf = (float)(o * feedback);
			pbuf++;
		} while(--c);
	}
	
	static void DoWorkNoOutput(float *pin, float *pbuf, int c, double const feedback, CBWState &s)
	{
		do
		{
	
			double delay = *pbuf;
	
			double const o = s.a[0] * delay + s.a[1] * s.i[0] + s.a[2] * s.i[1] - s.a[3] * s.o[0] - s.a[4] * s.o[1];
			s.o[1] = s.o[0];
			s.o[0] = (float)o;
			s.i[1] = s.i[0];
			s.i[0] = (float)delay;
			
			double dry = *pin++;
			*pbuf++ = (float)(feedback * o + dry);
	
		} while(--c);
	}
	
	static void DoWorkNoOutputNoFB(float *psamples, float *pbuf, int c)
	{
		memcpy(pbuf, psamples, c*4);
	}
	
	
	#pragma optimize ("", on)
	
	
	void WorkTrack(CTrack *pt, float *pin, float *pout, int numsamples, int const mode)
	{
		do
		{
			int count = min(numsamples, pt->Length - pt->Pos);
	
			if (count > 0)
			{
				if (mode == WM_NOIO)
				{
					if (pt->Feedback != 0)
						DoWorkNoInputNoOutput(pt->Buffer + pt->Pos, count, pt->Feedback, pt->Filter);
				}
				else if (mode == WM_WRITE)
				{
					if (pt->Feedback != 0)
						DoWorkNoInput(pout, pt->Buffer + pt->Pos, count, pt->WetOut, pt->Feedback, pt->Filter);
					else
						DoWorkNoInputNoFB(pout, pt->Buffer + pt->Pos, count, pt->WetOut, pt->Filter);
				}
				else if (mode == WM_READ)
				{
					if (pt->Feedback != 0)
						DoWorkNoOutput(pin, pt->Buffer + pt->Pos, count, pt->Feedback, pt->Filter);
					else
						DoWorkNoOutputNoFB(pin, pt->Buffer + pt->Pos, count);
	
				}
				else
				{
					if (pt->Feedback != 0)
						DoWork(pin, pout, pt->Buffer + pt->Pos, count, pt->WetOut, pt->Feedback, pt->Filter);
					else
						DoWorkNoFB(pin, pout, pt->Buffer + pt->Pos, count, pt->WetOut, pt->Filter);
				}
				
				pin += count;
				pout += count;
				numsamples -= count;
				pt->Pos += count;
			} 
	
			if (pt->Pos == pt->Length)
				pt->Pos = 0;
	
		} while(numsamples > 0);
	
	}
	float paux[MAX_BUFFER_LENGTH];
	bool Work(float *psamples, int numsamples, int const mode, CTrack* workTracks)
	{
		if (mode & WM_READ)
		{
			IdleMode = false;
			IdleCount = 0;
		}
		else
		{
			if (IdleMode)
			{
				return false;
			}
			else
			{
				IdleCount += numsamples;
				if (IdleCount >= DelayTime + MAX_BUFFER_LENGTH)
				{
					// clear all buffers
					for (int c = 0; c < track_count; c++)
						memset(workTracks[c].Buffer, 0, workTracks[c].Length*4);
					IdleMode = true;
				}
			}
		}
	
		//float *paux = pCB->GetAuxBuffer();
	
		if (mode & WM_READ)
			memcpy(paux, psamples, numsamples*4);
	
		if (!DryThru || !(mode & WM_READ))
			memset(psamples, 0, numsamples*4);
	
		for (int c = 0; c < track_count; c++)
			WorkTrack(&workTracks[c], paux, psamples, numsamples, mode);
	
		return true;
	}
	

private:

	void TickTrack(CTrack *pt, const lunar_track_t& ptval) {
		bool recalcfilter = false;
	
		if (ptval.filtertype)
		{
			pt->FilterType = *ptval.filtertype;
			recalcfilter = true;
		}
	
		if (ptval.filtercutoff)
		{
			pt->FilterCutoff = *ptval.filtercutoff;
			recalcfilter = true;
		}
	
		if (ptval.filterbw)
		{
			pt->FilterBandwidth = *ptval.filterbw;
			recalcfilter = true;
		}
	
		if (recalcfilter)
			SetFilter(*pt);
	
		if (ptval.unit)
			pt->Unit = *ptval.unit;
	
		if (ptval.length)
		{
			switch(pt->Unit)
			{
			case UNIT_MS:
				pt->Length = (int)(transport->samples_per_second * (*ptval.length / 1000.0));
				if (pt->Length < 1)
					pt->Length = 1;
				break;
			case UNIT_SAMPLE:
				pt->Length = *ptval.length;
				break;
			case UNIT_TICK:
				pt->Length = transport->samples_per_tick * *ptval.length;
				break;
			case UNIT_256:
				pt->Length = int(transport->samples_per_tick * *ptval.length) >> 8;
				if (pt->Length < 1)
					pt->Length = 1;
				break;
			}
	
		}
	
		if (pt->Length > MaxDelay)
		{
			pt->Length = MaxDelay;
		}
		
	
		if (pt->Pos >= pt->Length)
			pt->Pos = 0;
	
		if (ptval.feedback)
		{
			pt->Feedback = (float)((*ptval.feedback - 64) * (1.0 / 64.0)); 
		}
	
		if (ptval.wetout) {
			pt->WetOut = (float)(*ptval.wetout * (1.0 / 128.0));
		}
		
	}

	void SetFilter(CTrack &t)
	{
		switch(t.FilterType)
		{
		case FILTER_LP:	DSP_BW_InitLowpass(t.Filter, (float)t.FilterCutoff); break;
		case FILTER_HP:	DSP_BW_InitHighpass(t.Filter, (float)t.FilterCutoff); break;
		case FILTER_BP:	DSP_BW_InitBandpass(t.Filter, (float)t.FilterCutoff, (float)t.FilterBandwidth); break;
		case FILTER_BR:	DSP_BW_InitBandreject(t.Filter, (float)t.FilterCutoff, (float)t.FilterBandwidth); break;
		}
	
	}
		
	void InitTrack(CTrack& t)
	{
		delete[] t.Buffer;
		t.Buffer = new float [MaxDelay];
		memset(t.Buffer, 0, MaxDelay*4);
		t.Pos = 0;
		if (t.Length > MaxDelay)
			t.Length = MaxDelay;
	}

private:
	int MaxDelay;	// in samples
	int IdleCount;	// in samples
	int DelayTime;
	bool IdleMode;
	bool DryThru;

private:
	CTrack TracksL[MAX_TAPS];
	CTrack TracksR[MAX_TAPS];

};

lunar_fx *new_fx() { return new mi(); } 

/*

these are not 100% re-implemented:

char const *mi::DescribeValue(int const param, int const value)
{
	static char txt[16];

	switch(param)
	{
	case 1:		// length
		return NULL;
		break;
	case 2:		// unit
		switch(value)
		{
		case 0: return "tick";
		case 1: return "ms";
		case 2: return "sample";
		case 3: return "tick/256";
		}
		break;
	case 3:		// feedback
		sprintf(txt, "%.1f%%", (double)(value-64) * (100.0 / 64.0));
		break;
	case 4:
		switch(value)
		{
		case FILTER_LP: return "lowpass"; 
		case FILTER_HP: return "highpass";
		case FILTER_BP: return "bandpass";
		case FILTER_BR: return "bandreject";
		}
		break;
	case 5:
		sprintf(txt, "%dHz", value);
		break;
	case 6:
		sprintf(txt, "%dcps", value);
		break;
	case 7:		// wetout
		sprintf(txt, "%.1f%%", (double)value * (100.0 / 128.0));
		break;
	default:
		return NULL;
	}

	return txt;
}



void mi::AttributesChanged()
{
	MaxDelay = (int)(pMasterInfo->SamplesPerSec * (aval.maxdelay / 1000.0));
	for (int c = 0; c < numTracks; c++)
		InitTrack(c);
}


void mi::SetNumTracks(int const n)
{
	if (numTracks < n)
	{
		for (int c = numTracks; c < n; c++)
			InitTrack(c);
	}
	else if (n < numTracks)
	{
		for (int c = n; c < numTracks; c++)
			ResetTrack(c);
	
	}
	numTracks = n;

}
*/
